#!/usr/bin/env python3
"""
Spherical particles embedded in the middle of the layer on top of substrate.
"""
import bornagain as ba
from bornagain import ba_plot as bp, deg, nm, R3


def get_sample():
    """
    Returns a sample with spherical particles in a layer
    between vacuum and substrate.
    """

    # Define materials
    material_IntermLayer = ba.RefractiveMaterial("IntermLayer", 3.45e-06,
                                                 5.24e-09)
    material_Particle = ba.RefractiveMaterial("Particle", 0, 0)
    material_Substrate = ba.RefractiveMaterial("Substrate", 7.43e-06,
                                               1.72e-07)
    material_Vacuum = ba.RefractiveMaterial("Vacuum", 0, 0)

    # Define form factors
    ff = ba.Sphere(10.2*nm)

    # Define particles
    particle = ba.Particle(material_Particle, ff)
    particle_position = R3(0, 0, -25.2*nm)
    particle.translate(particle_position)

    # Define particle layouts
    layout = ba.ParticleLayout()
    layout.addParticle(particle)
    layout.setTotalParticleSurfaceDensity(0.01)

    # Define layers
    layer_1 = ba.Layer(material_Vacuum)
    layer_2 = ba.Layer(material_IntermLayer, 30*nm)
    layer_2.addLayout(layout)
    layer_3 = ba.Layer(material_Substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)
    sample.addLayer(layer_3)

    return sample


def get_simulation(sample):
    beam = ba.UnitBeam(0.15*nm, 0.15*deg)
    detector = ba.SphericalDetector(bp.simargs['n'], 2*deg, 0, 1*deg)
    simulation = ba.ScatteringSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    bp.parse_args(sim_n=200)
    sample = get_sample()
    simulation = get_simulation(sample)
    result = simulation.simulate()
    bp.plot_simulation_result(result)
