//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/InstrumentItems.h
//! @brief     Defines class InstrumentItem and all its children
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DEVICE_INSTRUMENTITEMS_H
#define BORNAGAIN_GUI_MODEL_DEVICE_INSTRUMENTITEMS_H

#include "Device/Coord/ICoordSystem.h"
#include "GUI/Model/CatDevice/BackgroundItemCatalog.h"
#include "GUI/Model/CatDevice/DetectorItemCatalog.h"
#include "GUI/Model/Descriptor/AxisProperty.h"
#include "GUI/Model/Descriptor/SelectionProperty.h"
#include "GUI/Model/Descriptor/VectorProperty.h"
#include "GUI/Model/Device/BackgroundItems.h"
#include "GUI/Model/Device/BeamItems.h"
#include "GUI/Model/Device/DetectorItems.h"
#include <functional>
#include <memory>

class BackgroundItem;
class DataItem;
class DepthprobeSimulation;
class ISimulation;
class MaskContainerItem;
class MultiLayer;
class OffspecSimulation;
class RealItem;
class ScatteringSimulation;

//! Abstract base class for instrument-specific item classes.

class InstrumentItem {
public:
    InstrumentItem();
    virtual ~InstrumentItem() = default;

    //! Creates an exact copy; also ID is the same!
    InstrumentItem* createItemCopy() const;

    virtual void writeTo(QXmlStreamWriter* w) const;
    virtual void readFrom(QXmlStreamReader* r);

    QString id() const { return m_id; }
    void setId(const QString& id) { m_id = id; }

    QString instrumentName() const { return m_name; }
    void setInstrumentName(const QString& instrumentName) { m_name = instrumentName; }

    //! The type as how to show it on the UI. Do not use for type checking or similar!
    virtual QString instrumentType() const = 0;

    template <typename T>
    bool is() const
    {
        return dynamic_cast<const T*>(this) != nullptr;
    }

    QString description() const { return m_description; }
    void setDescription(const QString& description) { m_description = description; }

    virtual BeamItem* beamItem() const;

    template <typename T>
    T* setBackgroundItemType();
    SelectionProperty<BackgroundItemCatalog>& backgroundSelection() { return m_background; }
    BackgroundItem* backgroundItem() const { return m_background.currentItem(); }

    virtual std::vector<int> shape() const = 0;

    virtual void importMasks(const MaskContainerItem*) {}

    virtual void updateToRealData(const RealItem* item) = 0;
    virtual bool alignedWith(const RealItem* item) const;

    //! The default user visible name when creating an instrument
    virtual QString defaultName() const = 0;

    virtual const ICoordSystem* createCoordSystem() const = 0;

    bool withPolarizerAnalyzer() const;
    void setWithPolarizerAnalyzer(bool with);

    VectorProperty& polarization() { return m_polarization; }
    const VectorProperty& polarization() const { return m_polarization; }
    void setPolarization(const R3& v) { m_polarization.setR3(v); }

    VectorProperty& analyzerDirection() { return m_analyzerDirection; }
    const VectorProperty& analyzerDirection() const { return m_analyzerDirection; }
    void setAnalyzerDirection(const R3& v) { m_analyzerDirection.setR3(v); }

    DoubleProperty& analyzerEfficiency() { return m_analyzerEfficiency; }
    const DoubleProperty& analyzerEfficiency() const { return m_analyzerEfficiency; }
    void setAnalyzerEfficiency(double v) { m_analyzerEfficiency.setValue(v); }

    DoubleProperty& analyzerTotalTransmission() { return m_analyzerTotalTransmission; }
    const DoubleProperty& analyzerTotalTransmission() const { return m_analyzerTotalTransmission; }
    void setAnalyzerTotalTransmission(double v) { m_analyzerTotalTransmission.setValue(v); }

    bool isExpandInfo() const { return m_expandInfo; }
    void setExpandInfo(bool b) { m_expandInfo = b; }

    bool isExpandPolarizerAlanyzer() const { return m_expandPolarizerAlanyzer; }
    void setExpandPolarizerAlanyzer(bool b) { m_expandPolarizerAlanyzer = b; }

    bool isExpandEnvironment() const { return m_expandEnvironment; }
    void setExpandEnvironment(bool b) { m_expandEnvironment = b; }

    virtual ISimulation* createSimulation(const MultiLayer& sample) const = 0;

protected:
    explicit InstrumentItem(const QString& modelType);

    QString m_id;
    QString m_name;
    QString m_description;
    bool m_withPolarizerAnalyzer;
    SelectionProperty<BackgroundItemCatalog> m_background;
    std::unique_ptr<BeamItem> m_beamItem;

    VectorProperty m_polarization;
    VectorProperty m_analyzerDirection;

    DoubleProperty m_analyzerEfficiency;
    DoubleProperty m_analyzerTotalTransmission;
    bool m_expandInfo = true;
    bool m_expandPolarizerAlanyzer = true;
    bool m_expandEnvironment = true;
};


class SpecularInstrumentItem : public InstrumentItem {
public:
    SpecularInstrumentItem();

    ScanItem* beamItem() const override;

    std::vector<int> shape() const override;
    void updateToRealData(const RealItem* item) override;
    bool alignedWith(const RealItem* item) const override;
    QString defaultName() const override;
    QString instrumentType() const override;

    const ICoordSystem* createCoordSystem() const override;

    ISimulation* createSimulation(const MultiLayer& sample) const override;
};


class DepthprobeInstrumentItem : public InstrumentItem {
public:
    DepthprobeInstrumentItem();

    void writeTo(QXmlStreamWriter* w) const override;
    void readFrom(QXmlStreamReader* r) override;

    ScanItem* beamItem() const override;

    std::vector<int> shape() const override;
    void updateToRealData(const RealItem* item) override;
    QString defaultName() const override;
    QString instrumentType() const override;


    const ICoordSystem* createCoordSystem() const override;

    AxisProperty& zAxis() { return m_zAxis; }
    const AxisProperty& zAxis() const { return m_zAxis; }

    bool isExpandParameters() const { return m_expandParameters; }
    void setExpandParameters(bool b) { m_expandParameters = b; }

    ISimulation* createSimulation(const MultiLayer& sample) const override;

protected:
    mutable std::unique_ptr<const ICoordSystem> m_coosys;
    AxisProperty m_zAxis;
    bool m_expandParameters = true;
};


class Instrument2DItem : public InstrumentItem {
public:
    void writeTo(QXmlStreamWriter* w) const override;
    void readFrom(QXmlStreamReader* r) override;

    template <typename T>
    T* setDetectorItemType();
    DetectorItem* detectorItem() const { return m_detector.currentItem(); }
    SelectionProperty<DetectorItemCatalog>& detectorSelection() { return m_detector; }

    void importMasks(const MaskContainerItem* maskContainer) override;

    static bool isDetectorPropertyName(const QString& name);

    bool isExpandDetector() const { return m_expandDetector; }
    void setExpandDetector(bool b) { m_expandDetector = b; }

protected:
    Instrument2DItem();

    SelectionProperty<DetectorItemCatalog> m_detector;
    bool m_expandDetector = true;
};


class GISASInstrumentItem : public Instrument2DItem {
public:
    std::vector<int> shape() const override;
    void updateToRealData(const RealItem* item) override;
    QString defaultName() const override;
    QString instrumentType() const override;

    const ICoordSystem* createCoordSystem() const override;

    std::unique_ptr<IDetector> createDetector() const;
    ISimulation* createSimulation(const MultiLayer& sample) const override;
};


class OffspecInstrumentItem : public Instrument2DItem {
public:
    OffspecInstrumentItem();

    void writeTo(QXmlStreamWriter* w) const override;
    void readFrom(QXmlStreamReader* r) override;

    std::vector<int> shape() const override;
    void updateToRealData(const RealItem* item) override;
    QString defaultName() const override;
    QString instrumentType() const override;

    const ICoordSystem* createCoordSystem() const override;

    AxisProperty& alphaAxis() { return m_alphaAxis; }
    const AxisProperty& alphaAxis() const { return m_alphaAxis; }

    bool isExpandBeamParameters() const { return m_expandBeamParameters; }
    void setExpandBeamParameters(bool b) { m_expandBeamParameters = b; }

    ISimulation* createSimulation(const MultiLayer& sample) const override;

protected:
    AxisProperty m_alphaAxis;
    bool m_expandBeamParameters = true;
};

//  ************************************************************************************************
//  templated functions implementation
//  ************************************************************************************************

template <typename T>
T* InstrumentItem::setBackgroundItemType()
{
    m_background.setCurrentItem<T>();
    return dynamic_cast<T*>(m_background.currentItem());
}

template <typename T>
T* Instrument2DItem::setDetectorItemType()
{
    m_detector.setCurrentItem<T>();
    return dynamic_cast<T*>(m_detector.currentItem());
}

#endif // BORNAGAIN_GUI_MODEL_DEVICE_INSTRUMENTITEMS_H
