//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/OffspecSimulation.cpp
//! @brief     Implements class OffspecSimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Simulation/OffspecSimulation.h"
#include "Base/Axis/Bin.h"
#include "Base/Axis/IAxis.h"
#include "Base/Pixel/IPixel.h"
#include "Base/Util/Assert.h"
#include "Device/Beam/Beam.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Data/Datafield.h"
#include "Device/Detector/IDetector.h"
#include "Device/Detector/SphericalDetector.h"
#include "Device/Histo/SimulationResult.h"
#include "Param/Distrib/DistributionHandler.h"
#include "Param/Distrib/Distributions.h"
#include "Resample/Element/DiffuseElement.h"
#include "Sim/Background/IBackground.h"
#include "Sim/Computation/DWBAComputation.h"

OffspecSimulation::OffspecSimulation(const Beam& beam, const MultiLayer& sample,
                                     const SphericalDetector& detector)
    : ISimulation(sample)
    , m_beam(beam.clone())
    , m_detector(detector.clone())
{
    m_detector->setDetectorNormal(m_beam->ki());
}

OffspecSimulation::OffspecSimulation(const MultiLayer& sample)
    : ISimulation(sample)
    , m_beam(Beam::horizontalBeam().clone())
    , m_detector(new SphericalDetector())
{
}

OffspecSimulation::~OffspecSimulation() = default;

std::vector<const INode*> OffspecSimulation::nodeChildren() const
{
    std::vector<const INode*> result = ISimulation::nodeChildren();
    result.push_back(m_beam.get());
    if (m_detector)
        result.push_back(m_detector.get());
    return result;
}

void OffspecSimulation::prepareSimulation()
{
    ASSERT(m_detector->rank() == 2);
    ASSERT(m_alpha_i_axis);
    ASSERT(m_alpha_i_axis->size() >= 1);
    m_active_indices = m_detector->active_indices();
    m_pixels.reserve(m_active_indices.size());
    for (auto detector_index : m_active_indices)
        m_pixels.emplace_back(m_detector->createPixel(detector_index));
}

void OffspecSimulation::setBeamParameters(double wavelength, const IAxis& alpha_axis, double phi_i)
{
    m_alpha_i_axis.reset(alpha_axis.clone());
    if (alpha_axis.size() < 1)
        throw std::runtime_error("OffspecSimulation::prepareSimulation() "
                                 "-> Error. Incoming alpha range size < 1.");
    beam() = InBeam(beam().intensity(), wavelength, alpha_axis.min(), phi_i);
}

const IAxis* OffspecSimulation::beamAxis() const
{
    return m_alpha_i_axis.get();
}

const ICoordSystem* OffspecSimulation::simCoordSystem() const
{
    ASSERT(beamAxis());
    return m_detector->offspecCoords(beamAxis()->clone());
}

//... Overridden executors:

//! init callbacks for setting the parameter values
void OffspecSimulation::initDistributionHandler()
{
    for (const auto& distribution : distributionHandler().paramDistributions()) {

        switch (distribution.whichParameter()) {
        case ParameterDistribution::BeamAzimuthalAngle:
            distributionHandler().defineCallbackForDistribution(
                &distribution, [&](double d) { m_beam->setAzimuthalAngle(d); });
            break;
        case ParameterDistribution::BeamInclinationAngle:
            distributionHandler().defineCallbackForDistribution(
                &distribution, [&](double d) { m_beam->setInclination(d); });
            break;
        case ParameterDistribution::BeamWavelength:
            distributionHandler().defineCallbackForDistribution(
                &distribution, [&](double d) { m_beam->setWavelength(d); });
            break;
        default:
            ASSERT(false);
        }
    }
}

void OffspecSimulation::runComputation(const ReSample& re_sample, size_t i, double weight)
{
    if (m_cache.empty())
        m_cache.resize(nElements(), 0.0);

    size_t j = i / m_active_indices.size();
    size_t k = i % m_active_indices.size();

    Beam beam2 = beam();
    beam2.setInclination(m_alpha_i_axis->bin(j).center());

    DiffuseElement ele(beam2.wavelength(), beam2.alpha_i(), beam2.phi_i(),
                       std::unique_ptr<IPixel>(m_pixels[k]->clone()), beam2.polMatrix(),
                       m_detector->analyzer().matrix(),
                       m_active_indices[k] == m_detector->indexOfSpecular(beam2));

    double intensity = DWBAComputation(re_sample, options(), progress()).simulatedIntensity(ele);

    double sin_alpha_i = std::abs(std::sin(ele.alphaI()));
    if (sin_alpha_i == 0.0) {
        intensity = 0;
    } else {
        const double solid_angle = ele.solidAngle();
        intensity *= m_beam->intensity() * solid_angle / sin_alpha_i;
    }

    if (background())
        intensity = background()->addBackground(intensity);

    m_cache[i] += intensity * weight;
}

//... Overridden getters:

bool OffspecSimulation::force_polarized() const
{
    return m_detector->analyzer().analyzerDirection() != R3{};
}

size_t OffspecSimulation::nElements() const
{
    return m_active_indices.size() * m_alpha_i_axis->size();
}

SimulationResult OffspecSimulation::packResult()
{

    // update intensity map
    ASSERT(m_alpha_i_axis);
    Datafield intensity_map({m_alpha_i_axis->clone(), m_detector->axis(1).clone()});
    intensity_map.setAllTo(0.);

    size_t ny = m_detector->axis(1).size();

    // Normalize, apply detector resolution and transfer detector image corresponding to
    // alpha_i = m_alpha_i_axis->bin(index)
    for (size_t j = 0; j < m_alpha_i_axis->size(); ++j) {
        Datafield detector_image({m_detector->axis(0).clone(), m_detector->axis(1).clone()});
        size_t N = detector_image.size();
        for (size_t i = 0; i < N; ++i)
            detector_image[i] = m_cache[j * N + i];
        m_detector->applyDetectorResolution(&detector_image);
        for (size_t i = 0; i < N; ++i)
            intensity_map[j * ny + i % ny] += detector_image[i];
    }

    return {intensity_map, simCoordSystem()};
}
