#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/Frame.h"
#include "Device/IO/IOFactory.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/JobItemUtils.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Model/Device/AxesItems.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Model/JobModel.h"
#include "GUI/Model/Model/RealModel.h"
#include "GUI/Support/IO/ImportDataInfo.h"
#include "GUI/Support/IO/ProjectUtils.h"
#include "GUI/Util/Error.h"
#include "GUI/Util/MessageService.h"
#include "GUI/Util/Path.h"
#include "Tests/GTestWrapper/google_test.h"
#include "Tests/Unit/GUI/Utils.h"
#include <QTest>
#include <memory>

class TestJobRealModel : public ::testing::Test {
protected:
    Datafield m_data{{new FixedBinAxis("x", 10, 0.0, 1.0), new FixedBinAxis("y", 10, -1.0, 1.0)}};
};


//! Test methods for retrieving nonXML data.

TEST_F(TestJobRealModel, nonXMLData)
{
    RealModel realModel;
    JobModel jobModel;

    // initial state
    EXPECT_EQ(realModel.dataItems().size(), 0);
    EXPECT_EQ(jobModel.dataItems().size(), 0);

    // adding RealItem
    auto* realData = realModel.insertIntensityDataItem();
    realData->setDatafield(UTest::GUI::createData().release());
    EXPECT_EQ(realModel.dataItems().size(), 1);

    // adding JobItem
    auto* jobItem = jobModel.createJobItem();
    auto* instrumentItem = new GISASInstrumentItem();
    jobItem->copyInstrumentIntoJob(instrumentItem);
    jobItem->createSimulatedDataItem();
    EXPECT_EQ(jobModel.dataItems().size(), 1);

    // adding RealItem to JobItem
    auto* realItem2 = jobItem->createRealItem();
    EXPECT_EQ(jobModel.dataItems().size(), 1);
    realItem2->setDatafield(UTest::GUI::createData(0.0, UTest::GUI::DIM::D1).release());
    EXPECT_EQ(jobModel.dataItems().size(), 2);
    EXPECT_EQ(realModel.dataItems().size(), 1);

    // checking data items of JobModel
    EXPECT_EQ(jobModel.dataItems().indexOf(jobItem->simulatedDataItem()), 0);
    EXPECT_EQ(jobModel.dataItems().indexOf(realItem2->dataItem()), 1);

    // Replacing the data inside RealItem with the data of the same dimensions
    realData->setDatafield(UTest::GUI::createData(2.0).release());
    EXPECT_EQ(realModel.dataItems().size(), 1);

    // Replacing the data inside RealItem with the data of different dimensions
    auto data = UTest::GUI::createData(3.0, UTest::GUI::DIM::D1);
    EXPECT_THROW(dynamic_cast<RealItem*>(realData)->setDatafield(data.get()), Error);

    EXPECT_EQ(realModel.dataItems().size(), 1);
}

//! Testing saving abilities of RealModel.

TEST_F(TestJobRealModel, RealModelWithData)
{
    const QString projectDir("test_DatafieldIOService");
    UTest::GUI::create_dir(projectDir);

    const double value1(1.0), value2(2.0), value3(3.0);

    RealModel realModel;
    RealItem* realItem1 = UTest::GUI::createRealData("data1", realModel, value1);
    RealItem* realItem2 = UTest::GUI::createRealData("data2", realModel, value2);

    realItem1->updateDataFileName();
    realItem2->updateDataFileName();

    // Saving first time
    QXmlStreamWriter w;
    realModel.writeTo(&w, projectDir);
    QTest::qSleep(10);

    // Checking existence of data on disk
    QString fname1 = "./" + projectDir + "/realdata_data1_0.int.gz";
    QString fname2 = "./" + projectDir + "/realdata_data2_0.int.gz";
    EXPECT_TRUE(QFile::exists(fname1));
    EXPECT_TRUE(QFile::exists(fname2));

    // Reading data from disk, checking it is the same
    std::unique_ptr<Datafield> dataOnDisk1(IOFactory::readDatafield(fname1.toStdString()));
    std::unique_ptr<Datafield> dataOnDisk2(IOFactory::readDatafield(fname2.toStdString()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *realItem1->dataItem()->datafield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *realItem2->dataItem()->datafield()));

    // Modifying data and saving the project.
    realItem2->setDatafield(UTest::GUI::createData(value3).release());
    realModel.writeTo(&w, projectDir);
    QTest::qSleep(10);

    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *realItem1->dataItem()->datafield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *realItem2->dataItem()->datafield()) == false);

    // checking that data on disk has changed
    dataOnDisk2.reset(IOFactory::readDatafield(fname2.toStdString()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *realItem2->dataItem()->datafield()));

    // Renaming RealData and check that file on disk changed the name
    realItem2->setRealItemName("data2new");
    realItem2->updateDataFileName();
    realModel.writeTo(&w, projectDir);
    QTest::qSleep(10);

    QString fname2new = "./" + projectDir + "/realdata_data2new_0.int.gz";
    EXPECT_TRUE(QFile::exists(fname2new));
    EXPECT_TRUE(UTest::GUI::isTheSame(fname2new, *realItem2->dataItem()->datafield()));

    // Check that file with old name was removed.
    EXPECT_FALSE(QFile::exists(fname2));
}

TEST_F(TestJobRealModel, RealItemWithNativeData)
{
    RealModel realModel;
    JobModel jobModel;

    // initial state
    auto dataItems = realModel.dataItems();
    EXPECT_EQ(dataItems.size(), 0);

    // adding RealItem
    auto* realItem = realModel.insertIntensityDataItem();
    EXPECT_EQ(realModel.dataItems().size(), 1);

    ImportDataInfo import_data(std::unique_ptr<Datafield>(m_data.clone()), "nbins");
    realItem->setImportData(std::move(import_data));
    // TODO: setImportData tested here with 2D data while main code uses it for 1D only.

    EXPECT_TRUE(realItem->isIntensityData());
    EXPECT_EQ(realModel.dataItems().size(), 2);
    realItem->setRealItemName("RealItem");

    // adding JobItem
    auto* jobItem = jobModel.createJobItem();
    jobItem->setIdentifier(QUuid::createUuid().toString());

    auto* instrumentItem = new GISASInstrumentItem();
    jobItem->copyInstrumentIntoJob(instrumentItem);
    jobItem->createSimulatedDataItem();
    EXPECT_EQ(jobModel.dataItems().size(), 1);

    // copying RealItem to JobItem
    jobItem->copyRealItemIntoJob(realItem);
    EXPECT_EQ(jobModel.dataItems().size(), 3);

    const QString projectDir("test_NativeData");
    UTest::GUI::create_dir(projectDir);

    // Saving
    QXmlStreamWriter w;
    realModel.writeTo(&w, projectDir);
    jobModel.writeTo(&w, projectDir);
    QTest::qSleep(10);

    // Checking existance of data on disk
    auto* data1 = realItem->dataItem();
    auto* data2 = realItem->nativeDataItem();
    auto* data3 = jobItem->realItem()->dataItem();
    auto* data4 = jobItem->realItem()->nativeDataItem();

    QString fname1 = "./" + projectDir + "/" + data1->fileName();
    QString fname2 = "./" + projectDir + "/" + data2->fileName();
    QString fname3 = "./" + projectDir + "/" + data3->fileName();
    QString fname4 = "./" + projectDir + "/" + data4->fileName();

    EXPECT_TRUE(QFile::exists(fname1));
    EXPECT_TRUE(QFile::exists(fname2));
    EXPECT_TRUE(QFile::exists(fname3));
    EXPECT_TRUE(QFile::exists(fname4));

    auto readData = [](const QString& filename) {
        return std::unique_ptr<Datafield>(IOFactory::readDatafield(filename.toStdString()));
    };

    // Reading data from disk, checking it is the same
    auto dataOnDisk1 = readData(fname1);
    auto dataOnDisk2 = readData(fname2);
    auto dataOnDisk3 = readData(fname3);
    auto dataOnDisk4 = readData(fname4);
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *data1->datafield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *data2->datafield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk3, *data3->datafield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk4, *data4->datafield()));
}
